@@ -54,19 +54,19 @@ class Callback(ABC):
     
     def on_train_begin(self):
         """ Called at the beginning of the training. """
-     
+        
     def on_train_end(self):
         """ Called at the end of the training. """
-      
+        
     def on_test_begin(self):
         """ Called at the beginning of the evaluation. """
-     
+        
     def on_test_end(self):
         """ Called at the end of the evaluation. """
     
     def on_epoch_begin(self, *args):
         """ Called at the beginning of an epoch. """
-
+        
     def on_epoch_end(self, *args):
         """ Called at the end of an epoch. """
     
@@ -97,7 +97,7 @@ class CallbackList(Callback):
     """
 
     callbacks: List[Callback]
-   
+    
     def on_train_begin(self, *args):
         for callback in self.callbacks:
             callback.on_train_begin(*args)
@@ -147,7 +147,7 @@ class TrainingStep(Callback):
     the input data) and a backward pass (where the obj:`optimizer`
     updates the obj:`model` weights and biases based on the difference
     between the predicted output and the expected output as expressed
-    by the obj:`loss_finction`).
+    by the obj:`loss_function`).
     """
     
     name: str = 'TrainingStep'
@@ -155,70 +155,67 @@ class TrainingStep(Callback):
     optimizer: tf.optimizers.Optimizer = None
     loss_function: tf.losses = None
     
-    def _train_step(self, features: InputFeatures):
+    def train_step(self, input_batch: InputFeatures):
         """ Train the obj:`model` on a single batch of examples. """
         
         with tf.GradientTape() as tape:
-            model_outputs = self.model.call(
-                input_ids=features.input_ids,
-                attention_mask=features.attention_mask,
-                token_type_ids=features.token_type_ids,
+            output_batch = self.model.call(
+                input_ids=input_batch.input_ids,
+                attention_mask=input_batch.attention_mask,
+                token_type_ids=input_batch.token_type_ids,
                 training=True,
             )
             
-            loss_values = self.loss_function(
-                features.label, model_outputs.logits
+            loss = self.loss_function(
+                input_batch.label, output_batch.logits
             )
             
-        trainable_variables = self.model.trainable_variables
-            
-        gradients = tape.gradient(loss_values, trainable_variables)
+        variables = self.model.trainable_variables
+        gradients = tape.gradient(loss, variables)
         
         self.optimizer.apply_gradients(
-            zip(gradients, trainable_variables)
-        )
-        
-        output_batch = TFSequenceClassifierBatchOutput(
-            loss=loss_values,
-            logits=model_outputs.logits,
-            hidden_states=model_outputs.hidden_states,
-            attentions=model_outputs.attentions,
-        )
-        
-        return output_batch
-    
-    def _test_step(self, features: InputFeatures):
-        """ Test the obj:`model` on a single batch of examples. """
-        
-        model_outputs = self.model.call(
-                  input_ids=features.input_ids,
-                  attention_mask=features.attention_mask,
-                  token_type_ids=features.token_type_ids,
-                  training=False,
-        )
-        
-        loss_values = self.loss_function(
-            features.label, model_outputs.logits
+            zip(gradients, variables)
         )
         
         return TFSequenceClassifierBatchOutput(
-            loss=loss_values,
-            logits=model_outputs.logits,
-            hidden_states=model_outputs.hidden_states,
-            attentions=model_outputs.attentions,
+            loss=loss,
+            logits=output_batch.logits,
+            hidden_states=output_batch.hidden_states,
+            attentions=output_batch.attentions,
+        )
+    
+    def test_step(self, input_batch: InputFeatures):
+        """ Test the obj:`model` on a single batch of examples. """
+        
+        output_batch = self.model.call(
+                  input_ids=input_batch.input_ids,
+                  attention_mask=input_batch.attention_mask,
+                  token_type_ids=input_batch.token_type_ids,
+                  training=False,
+        )
+        
+        loss = self.loss_function(
+            input_batch.label, output_batch.logits
+        )
+        
+        return TFSequenceClassifierBatchOutput(
+            loss=loss,
+            logits=output_batch.logits,
+            hidden_states=output_batch.hidden_states,
+            attentions=output_batch.attentions,
         )
     
     def on_train_begin(self):
         message = self.model.summary()
-        logger.info(message)           
-     
-    def on_train_batch_begin(self, step: int, features: InputFeatures):
-        self.train_step_output = self._train_step(features)
+        logger.info(message)
+        
+    def on_train_batch_begin(self, step: int, batch: InputFeatures):
+        self.train_step_output = self.train_step(batch)
              
-    def on_test_batch_begin(self, step: int, features: InputFeatures):
-        self.test_step_output = self._test_step(features)
-
+    def on_test_batch_begin(self, step: int, batch: InputFeatures):
+        self.test_step_output = self.test_step(batch)
 
+        
 @dataclass
 class History(Callback):
     """ Callback that records the performance of the model 
@@ -231,66 +228,84 @@ class History(Callback):
     
     name: str = 'History'
     epoch: int = field(default_factory=int)
-    metric: Callable[[], tf.metrics.Metric] = None
+    loss_metric: Callable[[], tf.metrics.Metric] = None
+    acc_metric: Callable[[], tf.metrics.Metric] = None
     training_step: Callback = None
-    train: Dict = field(default_factory=lambda: defaultdict(dict))
-    test: Dict = field(default_factory=lambda: defaultdict(dict))
-    train_details: Dict = field(default_factory=lambda: defaultdict(dict))
-    test_details: Dict = field(default_factory=lambda: defaultdict(dict))
-
+    logs: Dict = field(default_factory=lambda: defaultdict(dict))
     
     def __post_init__(self):
-        self.train_metric = self.metric()
-        self.test_metric = self.metric()
+        """ """
+        self.loss_train_metric = self.loss_metric()
+        self.loss_test_metric = self.loss_metric()
+        self.acc_train_metric = self.acc_metric()
+        self.acc_test_metric = self.acc_metric()
+        
+    def on_train_begin(self):
+        """ """
+        self.logs['loss'] = {}
+        self.logs['acc'] = {}
+        self.logs['val_loss'] = {}
+        self.logs['val_acc'] = {}
     
     def on_epoch_begin(self, epoch: int):
         """ Resets all of the metric state variables """
+        
         self.epoch = epoch
-        self.train_metric.reset_states()
-        self.test_metric.reset_states()
+        '''
+        self.logs['loss'][epoch] = []
+        self.logs['acc'][epoch] = []
+        self.logs['val_loss'][epoch] = []
+        self.logs['val_acc'][epoch] = []
+        '''
+        self.loss_train_metric.reset_states()
+        self.loss_test_metric.reset_states()
+        self.acc_train_metric.reset_states()
+        self.acc_test_metric.reset_states()
         
-        self.train['loss'] = {}
-        self.train_details['loss'] = {}
-        self.test_details['loss'] = {}
-        self.train['loss'][epoch] = []
-        self.train_details['loss'][epoch] = []
-        self.test_details['loss'][epoch] = []
-        
-        self.test['accuracy'] = {}
-        self.train_details['accuracy'] = {}
-        self.test_details['accuracy'] = {}
-        self.test['accuracy'][epoch] = []
-        self.train_details['accuracy'][epoch] = []
-        self.test_details['accuracy'][epoch] = []
-    
     def on_epoch_end(self, epoch: int):
         """ Display the performance metrics after each epoch """
-        self.train['loss'][self.epoch] = self.train_metric.result()
-        self.test['loss'][self.epoch] = self.test_metric.result()
         
-        message = f'Epoch {self.epoch+1:4d} {self.name:10} ' \
-                  f'Train {self.train_metric.result():1.3f} ' \
-                  f'Test {self.test_metric.result():1.3f}'
+        message = f"Training loss:       {self.logs['loss'][epoch]:1.3f}" \
+                  f"Training accuracy:   {self.logs['acc'][epoch]:1.3f}" \
+                  f"Validation loss:     {self.logs['val_loss'][epoch]:1.3f}" \
+                  f"Validation accuracy: {self.logs['val_acc'][epoch]:1.3f}"
         
-        logger.info(message)
+        logger.info(message)     
 
-    def on_train_batch_end(self, step: int):
-        """ Compute the performance metrics after each training batch """
-        loss = self.training_step.train_step_output.loss
-        self.train_metric(loss)
-        self.train_details['loss'][self.epoch].extend(loss)
-        self.train['loss'][self.epoch] = self.train_metric.result()
+    def on_train_batch_end(self, step: int, input_batch: InputFeatures):
+        """ Compute the performance metrics for each training batch """
+        epoch = self.epoch
+        output_batch = self.training_step.train_step_output
+
+        labels = input_batch.label
+        logits = output_batch.logits
+        loss = output_batch.loss
         
-    def on_test_batch_end(self, step: int):
-        """ Compute the performance metrics after each testing batch """
-        loss = self.training_step.test_step_output.loss
-        self.test_metric(loss)
-        self.test_details['loss'][self.epoch].extend(loss)
-        self.test['loss'][self.epoch] = self.test_metric.result()
+        self.loss_train_metric.update_state(loss)
+        self.acc_train_metric.update_state(labels, logits)
+        
+        self.logs['loss'][epoch] = self.loss_train_metric.result()
+        self.logs['acc'][epoch] = self.acc_train_metric.result()
+        
+        
+    def on_test_batch_end(self, step: int, input_batch: InputFeatures):
+        """ Compute the performance metrics for each testing batch """
+        epoch = self.epoch
+        output_batch = self.training_step.test_step_output
+        
+        labels = input_batch.label
+        logits = output_batch.logits
+        loss = output_batch.loss
+        
+        self.loss_test_metric.update_state(loss)
+        self.acc_test_metric.update_state(labels, logits)
+        
+        self.logs['val_loss'][epoch] = self.loss_test_metric.result()
+        self.logs['val_acc'][epoch] = self.acc_test_metric.result()
         
     @property
     def best_result(self) -> float:
-        return min(self.test['loss'].values())
+        return min(self.logs['val_loss'].values())
     
 
 @dataclass
@@ -321,7 +336,7 @@ class ModelCheckpoint(Callback):
 
     def on_epoch_end(self, epoch: int):
         """ Pass the obj:`ModelCheckpoint` callback after the obj:`History`. """
-        result = self.history.test['loss'][epoch]
+        result = self.history.logs['val_loss'][epoch]
         diff = self.best_result - result
         is_better = diff > 0 if self.direction == 'minimize' else diff < 0
         if is_better:
@@ -389,7 +404,7 @@ class EarlyStopping(Callback):
 
     def on_epoch_end(self, epoch: int):
         """ """
-        result = self.history.test['loss'][epoch]
+        result = self.history.logs['val_loss'][epoch]
         diff = self.best_result - result
         is_better = diff > self.min_delta if self.direction == 'minimize' \
             else diff < self.min_delta * -1
@@ -448,7 +463,7 @@ class ProgbarLogger(Callback):
     console. """ 
     
     name: str = 'ProgbarLogger'
-    history: List[History] = None
+    history: History = None
     metric: Callable = tf.metrics.Mean
     epoch: int = field(default_factory=int)
     num_samples: int = None
@@ -515,10 +530,10 @@ class ProgbarLogger(Callback):
             if self.show_epoch_progress and time_diff >= self.update_interval:
                 
                 self.epoch_progress_tqdm.desc = \
-                     f'loss: {self.train_logs["loss"]:04F} - ' \
-                     f'val_loss: {self.test_logs["loss"]:04F}'
-                     #f'acc: {self.logs["accuracy"]:04F} - ' \
-                     #f'val_acc: {self.logs["test_accuracy"]:04F}'
+                     f"loss: {self.train_logs['loss']:04F} - " \
+                     f"accuracy: {self.train_logs['acc']:04F} - " \
+                     f"val_loss: {self.test_logs['loss']:04F} - " \
+                     f"val_accuracy: {self.test_logs['acc']:04F}"
 
                 self.epoch_progress_tqdm.update(self.steps_to_update)
                 self.steps_to_update = 0
@@ -547,7 +562,7 @@ class ProgbarLogger(Callback):
     def on_test_end(self):
         if not self.is_training:
             self._clean_up_progbar("test")
-     
+            
     def on_epoch_begin(self, epoch: int):
         self.epoch = epoch
         self._initialize_progbar("train", epoch)
@@ -558,14 +573,16 @@ class ProgbarLogger(Callback):
             self.overall_progress_tqdm.update(1)
         self.epoch_progress_tqdm.close()
 
-    def on_train_batch_end(self, step: int):
+    def on_train_batch_end(self, step: int, input_batch: InputFeatures):
         train_logs = defaultdict(float)
-        train_logs["loss"] = self.history.train['loss'][self.epoch]
+        train_logs["loss"] = self.history.logs['loss'][self.epoch]
+        train_logs["acc"] = self.history.logs['acc'][self.epoch]
         self.train_logs = train_logs
         self._update_progbar()
         
-    def on_test_batch_end(self, step: int):
+    def on_test_batch_end(self, step: int, input_batch: InputFeatures):
         test_logs = defaultdict(float)
-        test_logs["loss"] = self.history.test['loss'][self.epoch]
+        test_logs["loss"] = self.history.logs['val_loss'][self.epoch]
+        test_logs["acc"] = self.history.logs['val_acc'][self.epoch]
         self.test_logs = test_logs 
         self._update_progbar()
