@@ -40,18 +40,21 @@ class Trainer():
             callbacks.on_train_begin()
             
             for epoch in range(self.num_epochs):
-                
                 callbacks.on_epoch_begin(epoch)
                 
                 for step, batch in enumerate(self.train_stream):
                     callbacks.on_train_batch_begin(step, batch)
-                    callbacks.on_train_batch_end(step)
+                    callbacks.on_train_batch_end(step, batch)
                     
                 if self.test_stream:
+                    callbacks.on_test_begin()
+                    
                     for step, batch in enumerate(self.test_stream):
                         callbacks.on_test_batch_begin(step, batch)
-                        callbacks.on_test_batch_end(step)
-                       
+                        callbacks.on_test_batch_end(step, batch)
+                        
+                    callbacks.on_test_end()
+                    
                 callbacks.on_epoch_end(epoch)
                 
             callbacks.on_train_end()
@@ -60,4 +63,5 @@ class Trainer():
             message = 'Unexpected error during the training process.'
             logger.error(message)
             exit()
+
             
\ No newline at end of file
